<?php
class Zip { 

function put_contents($file, $contents, $mode = false ) {
		if ( ! ($fp = @fopen($file, 'w')) )
			return false;
		@fwrite($fp, $contents);
		@fclose($fp);
		chmod($file, $mode);
		return true;
	}
	
function rtrail($string) {
	return rtrim($string, '/');
}

function mkdir($path, $chmod = false, $chown = false, $chgrp = false) {
		// safe mode fails with a trailing slash under certain PHP versions.
		$path = Zip::rtrail($path);
		if ( empty($path) )
			$path = '/';

		if ( ! $chmod )
			$chmod = 0755;

		if ( ! @mkdir($path) )
			return false;
			chmod($path, $chmod);
		if ( $chown )
			//chown($path, $chown);
		if ( $chgrp )
			//chgrp($path, $chgrp);
		return true;
	}
	
/**
 *
 * @param string $file Full path and filename of zip archive
 * @param string $to Full path on the filesystem to extract archive to
 * @return mixed WP_Error on failure, True on success
 */
function unzip_file($file, $to) {

	// Unzip can use a lot of memory, but not this much hopefully
	@ini_set('memory_limit', '256M');

	if ( class_exists('ZipArchive') ) {
			$result = Zip::unzip($file, $to );
			if ( true === $result ) {
				return $result;
			} 
		}
	
	// Fall through to PclZip if ZipArchive is not available, or encountered an error opening the file.
	echo Zip::pclzip($file, $to);
}

/**
 *
 * @param string $file Full path and filename of zip archive
 * @param string $to Full path on the filesystem to extract archive to
 * @param array $needed_dirs A partial list of required folders needed to be created.
 * @return mixed WP_Error on failure, True on success
 */
function unzip($file, $to ) {

	$z = new ZipArchive();

	// PHP4-compat - php4 classes can't contain constants
	$zopen = $z->open($file, /* ZIPARCHIVE::CHECKCONS */ 4);
	if ( true !== $zopen )
		return 'incompatible archive';


	for ( $i = 0; $i < $z->numFiles; $i++ ) {
		if ( ! $info = $z->statIndex($i) )
			return 'Could not retrieve file from archive.';

		if ( '/' == substr($info['name'], -1) ) // directory
			continue;

		if ( '__MACOSX/' === substr($info['name'], 0, 9) ) // Don't extract the OS X-created __MACOSX directory files
			continue;

		$contents = $z->getFromIndex($i);
		if ( false === $contents )
			return 'Could not extract file from archive.';

		if ( ! Zip::put_contents( $to . $info['name'], $contents, 0644) )
			return 'Could not copy file.';
	}

	$z->close();

	return true;
}

/**
 *
 * @param string $file Full path and filename of zip archive
 * @param string $to Full path on the filesystem to extract archive to
 * @param array $needed_dirs A partial list of required folders needed to be created.
 * @return mixed WP_Error on failure, True on success
 */
function pclzip($file, $to) {

	require_once(ROOT . 'ip-inc/lib/pclzip.lib.php');

	$archive = new PclZip($file);
	
	// Is the archive valid?
	if ( false == ($archive_files = $archive->extract(PCLZIP_OPT_EXTRACT_AS_STRING)) )
		return 'Incompatible Archive.';

	if ( 0 == count($archive_files) )
		return 'Empty archive.';

	// Determine any children directories needed (From within the archive)
	foreach ( $archive_files as $file ) {
		if ( '__MACOSX/' === substr($file['filename'], 0, 9) ) // Skip the OS X-created __MACOSX directory
			continue;

		$needed_dirs[] = $to . Zip::rtrail( $file['folder'] ? $file['filename'] : dirname($file['filename']) );
	}

	$needed_dirs = array_unique($needed_dirs);
	foreach ( $needed_dirs as $dir ) {
		// Check the parent folders of the folders all exist within the creation array.
		if ( Zip::rtrail($to) == $dir ) // Skip over the working directory, We know this exists (or will exist)
			continue;
		if ( strpos($dir, $to) === false ) // If the directory is not within the working directory, Skip it
			continue;

		$parent_folder = dirname($dir);
		while ( !empty($parent_folder) && Zip::rtrail($to) != $parent_folder && !in_array($parent_folder, $needed_dirs) ) {
			$needed_dirs[] = $parent_folder;
			$parent_folder = dirname($parent_folder);
		}
	}
	asort($needed_dirs);

	// Create those directories if need be:
	foreach ( $needed_dirs as $_dir ) {
		if ( ! Zip::mkdir($_dir, 0755) && ! is_dir($_dir) ) // Only check to see if the dir exists upon creation failure. Less I/O this way.
			return 'Could not create directory.';
	}
	unset($needed_dirs);

	// Extract the files from the zip
	foreach ( $archive_files as $file ) {
		if ( $file['folder'] )
			continue;

		if ( '__MACOSX/' === substr($file['filename'], 0, 9) ) // Don't extract the OS X-created __MACOSX directory files
			continue;

		if ( ! Zip::put_contents( $to . $file['filename'], $file['content'], 0644) )
			return 'Could not copy file.';
	}
	return true;
}


}